/**
 * Copyright (C) 2003-2004 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

package sync4j.exchange.items.note.dao;

import java.io.IOException;
import java.text.MessageFormat;
import java.util.ArrayList;

import sync4j.exchange.httptransport.WebDavHttpTransport;
import sync4j.exchange.items.note.model.Note;
import sync4j.exchange.items.common.dao.ItemDAO;
import sync4j.exchange.AuthenticationException;
import sync4j.exchange.ExchangeAccessException;
import sync4j.exchange.xml.XmlParseException;
import sync4j.exchange.xml.XmlParser;

import sync4j.exchange.DataAccessException;

import sync4j.framework.server.store.NotFoundException;

/*
 * This class implements methods to access note data
 * in exchange server datastore
 *
 * @author Fabio Maggi @ Funambol
 * @version $Id: NoteDAO.java,v 1.2 2004/04/06 11:26:18 ste Exp $
 *
 **/
public class NoteDAO  extends ItemDAO {

    //---------------------------------------------------------------- Constants

    private static final String WEBDAV_MSG_PROPPATCH_UPDATE_NOTE             =
    "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n"                          +
    "<D:propertyupdate xmlns:D=\"DAV:\" "                                    +
    "xmlns:EX=\"http://schemas.microsoft.com/exchange/\" "                   +
    "xmlns:HN=\"urn:schemas:httpmail:\">\n"                                  +
    "<D:set>\n"                                                              +
    " <D:prop>\n"                                                            +
    "  <D:contentclass>urn:content-classes:person</D:contentclass>\n"        +
    "  <EX:outlookmessageclass>IPM.StickyNote</EX:outlookmessageclass>\n"    +
    "  <HN:subject>{0}</HN:subject>\n"                                       +
    "  <HN:textdescription>{1}</HN:textdescription>\n"                       +
    "  <HN:date>{2}</HN:date>\n"                                             +
    " </D:prop>\n"                                                           +
    "</D:set>\n"                                                             +
    "</D:propertyupdate>"                                                    ;

    private static final String WEBDAV_MSG_SELECT_NOTES                      =
    "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n"                          +
    "<D:searchrequest xmlns:D =\"DAV:\">\n"                                  +
    "<D:sql>\n"                                                              +
    "Select "                                                                +
    "\"urn:schemas:httpmail:textdescription\" AS textdescription, "          +
    "\"urn:schemas:httpmail:subject\" AS subject, "                          +
    "\"urn:schemas:httpmail:date\" AS date, "                                +
    "\"http://schemas.microsoft.com/repl/repl-uid\" AS repluid, "            +
    "\"DAV:getlastmodified\" AS getlastmodified, "                           +
    "\"DAV:isfolder\" AS isfolder "                                          +
    "FROM \"/{0}/{1}/{2}\"  "                                                +
    "{3} "                                                                   +
    "</D:sql>\n"                                                             +
    "</D:searchrequest>"                                                     ;

    private static final String TAG_DATE               = "date"              ;
    private static final String TAG_DESCRIPTION        =
        "textdescription"                                                    ;
    private static final String TAG_RESPONSE           = "a:response"        ;
    private static final String TAG_SUBJECT            = "subject"           ;
    private static final String TAG_REPLUID            = "repluid"           ;
    private static final String TAG_LAST_MODIFIED      = "getlastmodified"   ;

    private static final String NEW_LINE               = "\n"                ;

    //------------------------------------------------------------- Private data

    private String               exchangeServerHost  = null ;
    private int                  exchangeServerPort  = 0    ;

    private String               timezone            = null ;
    private long                 timediff            = 0    ;

    private WebDavHttpTransport  webDavHttp          = null ;

    //------------------------------------------------------------- Costructors

    public NoteDAO(String exchangeServerHost ,
                   int    exchangeServerPort ,
                   String timezone)
        throws DataAccessException {

        this.exchangeServerHost = exchangeServerHost ;
        this.exchangeServerPort = exchangeServerPort ;
        this.timezone           = timezone           ;

        this.webDavHttp         = new WebDavHttpTransport(exchangeServerHost,
                                                          exchangeServerPort);

    }
    //------------------------------------------------------------- Public methods

    /**
     * insert / update note
     *
     * @param note
     * @param username
     * @param credentials
     * @param sourceURI
     *
     * @return new / updated note
     *
     * @throws sync4j.exchange.util.DataAccessException
     **/
    public Note setNote(Note    note        ,
                        String  username    ,
                        String  credentials ,
                        String  sourceURI)
    throws DataAccessException {

        String webDavMsg               = null ;

        String response                = null ;

        String id                      = null ;

        String webDavNoteMsgLength     = null ;
        String webDavNoteMsg           = null ;
        String webDavHeaderMsg         = null ;

        String date                    = null ;

        String server                  = null ;
        String resource                = null ;

        server   = getServerFromURI   (sourceURI );
        resource = getResourceFromURI (sourceURI );

        try {

            date = XmlParser.dateToWebDavTag(note.getDate());

        } catch (XmlParseException e) {
            throw new DataAccessException("Error parsing date item", e);
        }


        webDavNoteMsg
            = MessageFormat.format(WEBDAV_MSG_PROPPATCH_UPDATE_NOTE,
              new Object[] {
                note.getSubject()          ,
                note.getTextDescription()  ,
                date
              }
        );

        webDavNoteMsgLength = String.valueOf(webDavNoteMsg.length());

        webDavHeaderMsg
            = MessageFormat.format(ItemDAO.WEBDAV_HEADER_PROPPATCH,
                new Object[] {
                    "/"                  +
                        server           +
                        "/"              +
                        username         +
                        "/"              +
                        resource             ,
                    note.getHref()           ,
                    this.exchangeServerHost  ,
                    webDavNoteMsgLength      ,
                    credentials

              }
        );

        webDavMsg = webDavHeaderMsg  +
                    NEW_LINE         +
                    webDavNoteMsg    ;

        try {
            response = this.webDavHttp.sendRequest(webDavMsg);
        } catch (Exception e) {
            e.printStackTrace();
            throw new
                DataAccessException("Error get Exchange server response", e);
        }

        try {
            checkResponseStatus(response);
        } catch (Exception e) {

            throw new DataAccessException("USER "                   +
                                          username                  +
                                          " URI "                   +
                                          sourceURI                 +
                                          " get exchange note"      ,
                                          e)                        ;

        }

        try {
            id = XmlParser.getRuidFromResponse(response);
        } catch (XmlParseException e) {
            e.printStackTrace();
            throw new DataAccessException("Error get note id", e);
        }

        note.setId(id);

        return note;

    }

    /**
     * Delete a <i>Note</i> from Exchange server.
     *
     * @param note
     * @param username
     * @param credentials
     * @param sourceURI
     *
     * @throws sync4j.exchange.util.DataAccessException
     *
     */
    public void removeNote(Note    note     ,
                           String  username    ,
                           String  credentials ,
                           String  sourceURI)
        throws DataAccessException {

        String webDavNoteMsg           = null ;
        String webDavNoteMsgLength     = null ;
        String webDavHeaderMsg         = null ;
        String webDavMsg               = null ;
        String response                = null ;

        String server                  = null ;
        String resource                = null ;

        server   = getServerFromURI   (sourceURI );
        resource = getResourceFromURI (sourceURI );

        webDavNoteMsg          = ""                ;
        webDavNoteMsgLength    =
            String.valueOf(webDavNoteMsg.length()) ;

        webDavHeaderMsg =
            MessageFormat.format(ItemDAO.WEBDAV_HEADER_REMOVE,
            new Object[] {
                        "/"            +
                            server     +
                            "/"        +
                            username   +
                            "/"        +
                            resource             ,
                        note.getHref()           ,
                        this.exchangeServerHost  ,
                        webDavNoteMsgLength      ,
                        credentials

            }
        );

        webDavMsg = webDavHeaderMsg  +
                    NEW_LINE         +
                    webDavNoteMsg    ;

        try {
            response = this.webDavHttp.sendRequest(webDavMsg);
        } catch (IOException e) {
            throw new
                DataAccessException("Error get Exchange server response", e);
        }

        try {
            checkResponseStatus(response);
        } catch (Exception e) {

            throw new DataAccessException("USER "                   +
                                          username                  +
                                          " URI "                   +
                                          sourceURI                 +
                                          " remove exchange note"   ,
                                          e)                        ;

        }

    }

    /**
     * get notes from Exchange Server
     *
     * @param username
     * @param credentials
     * @param ids
     * @param sourceURI
     *
     * @return array of find notes
     *
     * @throws sync4j.exchange.util.DataAccessException
     **/
    public Note[] getNotes (String    username    ,
                            String    credentials ,
                            String[]  ids         ,
                            String    sourceURI)
        throws DataAccessException {


        Note[] exchangeNotes          = null ;

        String response               = null ;
        String webDavNoteMsg          = null ;
        String webDavNoteMsgLength    = null ;
        String webDavMsg              = null ;
        String webDavHeaderMsg        = null ;

        String clause                 = null ;

        String server                 = null ;
        String resource               = null ;

        server   = getServerFromURI   (sourceURI );
        resource = getResourceFromURI (sourceURI );

        clause = XmlParser.getClause(ids);

        webDavNoteMsg =
            MessageFormat.format(WEBDAV_MSG_SELECT_NOTES,
            new Object[] {server, username, resource, clause});

        webDavNoteMsgLength
            = String.valueOf(webDavNoteMsg.length());

        webDavHeaderMsg =
            MessageFormat.format(ItemDAO.WEBDAV_HEADER_SELECT,

            new Object[] {
                        "/"                     +
                            server              +
                            "/"                 +
                            username            +
                            "/"                 +
                            resource              ,
                        this.exchangeServerHost   ,
                        webDavNoteMsgLength       ,
                        credentials

            }
        );

        webDavMsg = webDavHeaderMsg  +
                    NEW_LINE         +
                    webDavNoteMsg    ;

        try {
            response = this.webDavHttp.sendRequest(webDavMsg);
        } catch (IOException e) {
            throw new
                DataAccessException("Error get Exchange server response", e);
        }

        try {
            checkResponseStatus(response);
        } catch (Exception e) {

            throw new DataAccessException("USER "                   +
                                          username                  +
                                          " URI "                   +
                                          sourceURI                 +
                                          " get exchange note"      ,
                                          e)                        ;

        }

        exchangeNotes = getNotesFromWebDavMsg(response);

        return exchangeNotes;

    }

     //------------------------------------------------------------- Private methods

    /**
     * Create a <i>Note</i> array object from WebDav msg.
     *
     * @param id
     * @param webDavMsg
     *
     * @return a newly create <i>Note</i> object initialized with the
     *         fields in the result set
     *
     */
    private Note[] getNotesFromWebDavMsg(String webDavMsg)
    throws DataAccessException {

        ArrayList  notes      = null ;
        String []  resps      = null ;
        String []  msg        = null ;

        String     response   = null ;
        String     id         = null ;
        String     replUid    = null ;

        String     isFolder   = null ;
        int        count      = 0    ;

        msg   = new String[] {webDavMsg};

        notes = new ArrayList();

        try {

            resps = XmlParser.getXMLTag(msg, TAG_RESPONSE);

            for (int i = 0, l = resps.length; i < l; i++) {

                response =  resps[i]                                        ;
                replUid  =  XmlParser.getXMLTagValue(response, TAG_REPLUID) ;
                id       =  XmlParser.getIdFromReplUid(replUid)             ;

                isFolder    =  XmlParser.getXMLInitTagValue(response,
                                                            TAG_IS_FOLDER)  ;
                if(PROP_NO_FOLDER.equals(isFolder)) {
                    notes.add(getNoteFromResponseTag(id, response));
                    count ++;
                }
            }

        } catch (XmlParseException e) {
            throw new DataAccessException("Error parsing note item", e);
        }

        return (Note[])notes.toArray(new Note[count]);

    }

    /**
     * Create a <i>Note</i> object from response tag.
     *
     * @param id
     * @param msgResponse
     *
     * @return a newly create <i>Note</i> object initialized with the
     *         fields in the result set
     *
     */
    private Note getNoteFromResponseTag(String id, String msgResponse)
    throws DataAccessException {

        Note    n           = null ;
        String  lastUpdate  = null ;
        String  date        = null ;

        n = new Note(id);

        try {

            n.setTextDescription         (XmlParser.getXMLTagValue
                                            (msgResponse, TAG_DESCRIPTION))  ;

            n.setSubject                 (XmlParser.getXMLTagValue
                                            (msgResponse, TAG_SUBJECT))      ;

            date        =                 XmlParser.getXMLInitTagValue
                                            (msgResponse, TAG_DATE)          ;

            lastUpdate  =                 XmlParser.getXMLInitTagValue
                                            (msgResponse, TAG_LAST_MODIFIED) ;

            n.setDate                    (XmlParser.webDavTagToDate(date))   ;

            n.setLastModified            (XmlParser.webDavTagToDate
                                            (lastUpdate))                    ;

        } catch (XmlParseException e) {
            throw new DataAccessException("Error parsing note item", e);
        }

        return n;

    }

}